"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClickHouseDataSource = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@vulcan-sql/core");
const stream_1 = require("stream");
const fs = require("fs");
const client_1 = require("@clickhouse/client");
const sqlBuilder_1 = require("./sqlBuilder");
const typeMapper_1 = require("./typeMapper");
const lodash_1 = require("lodash");
let ClickHouseDataSource = class ClickHouseDataSource extends core_1.DataSource {
    constructor() {
        super(...arguments);
        this.logger = this.getLogger();
        this.clientMapping = new Map();
    }
    onActivate() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const profiles = this.getProfiles().values();
            for (const profile of profiles) {
                this.logger.debug(`Initializing profile: ${profile.name} using clickhouse driver`);
                // Omit and get the tls from connection, because clickhouse client support tls for bytes data,
                const tls = (0, lodash_1.omit)(profile.connection, ['tls']);
                const options = Object.assign({ application: 'VulcanSQL' }, profile.connection);
                // Set TLS options is existed
                if (tls.ca_cert && tls.cert && tls.key) {
                    options.tls = {
                        ca_cert: fs.readFileSync(tls.ca_cert),
                        cert: fs.readFileSync(tls.cert),
                        key: fs.readFileSync(tls.key),
                    };
                }
                else if (tls.ca_cert) {
                    options.tls = { ca_cert: fs.readFileSync(tls.ca_cert) };
                }
                const client = (0, client_1.createClient)(options);
                this.clientMapping.set(profile.name, {
                    client,
                    options,
                });
                yield client.query({ query: 'SELECT 1' });
                this.logger.debug(`Profile ${profile.name} initialized`);
            }
        });
    }
    execute({ statement: sql, bindParams, profileName, operations, }) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.checkProfileExist(profileName);
            const { client } = this.clientMapping.get(profileName);
            // convert to clickhouse support type of query params.
            const params = this.convertToQueryParams(bindParams);
            try {
                // Use JSONEachFormat to get data result with column name , refer: https://clickhouse.com/docs/en/integrations/language-clients/nodejs#supported-data-formats
                const builtSQL = (0, sqlBuilder_1.buildSQL)(sql, operations);
                const data = yield client.query({
                    query: builtSQL,
                    format: 'JSONEachRow',
                    query_params: params,
                });
                // Get the query metadata e.g: column name, type by DESCRIBE TABLE method.
                // DESCRIBE TABLE will return the type after evaluating the query, without execution, and will return only one entry per column, see: https://www.tinybird.co/blog-posts/tips-11-how-to-get-the-types-returned-by-a-query
                const metadata = yield client.query({
                    // remove semicolon at the end, because could not exist semicolon in sub-query when using DESCRIBE TABLE
                    query: `DESCRIBE TABLE (${(0, sqlBuilder_1.removeEndingSemiColon)(builtSQL)})`,
                    format: 'JSONEachRow',
                    query_params: params,
                });
                return yield this.getResultFromRestfulSet(metadata, data);
            }
            catch (e) {
                this.logger.debug(`Errors occurred, release connection from ${profileName}`);
                throw e;
            }
        });
    }
    prepare({ parameterIndex, value }) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // ClickHouse use {name:type} be a placeholder, so if we only use number string as name e.g: {1:Unit8}
            // it will face issue when converting to the query params => {1: value1}, because the key is value not string type, so here add prefix "p" to avoid this issue.
            return `{p${parameterIndex}:${(0, typeMapper_1.mapToClickHouseType)(value)}}`;
        });
    }
    destroy() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            for (const { client } of this.clientMapping.values()) {
                yield client.close();
            }
        });
    }
    getResultFromRestfulSet(metadata, data) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const dataRowStream = new stream_1.Stream.Readable({
                objectMode: true,
                read: () => null,
                // automatically destroy() the stream when it emits 'finish' or errors. Node > 10.16
                autoDestroy: true,
            });
            // Get the metadata and only need column name and type
            const columns = (yield metadata.json());
            const rawStream = data.stream();
            // ClickHouse stream only called once and return all data row in one chuck, so we need to push each row to the stream by loop.
            // Please see https://clickhouse.com/docs/en/integrations/language-clients/nodejs#resultset-and-row-abstractions
            rawStream.on('data', (rows) => {
                rows.forEach((row) => dataRowStream.push(JSON.parse(row.text)));
            });
            yield new Promise((resolve) => {
                rawStream.on('end', () => {
                    dataRowStream.push(null);
                    resolve(null);
                });
            });
            return {
                getColumns: () => {
                    return columns.map((column) => ({
                        name: column.name || '',
                        // Convert ClickHouse type to FieldDataType supported by VulcanSQL for generating the response schema in the specification, see: https://github.com/Canner/vulcan-sql/pull/78#issuecomment-1621532674
                        type: (0, typeMapper_1.mapFromClickHouseType)(column.type || ''),
                    }));
                },
                getData: () => dataRowStream,
            };
        });
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    export(options) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            throw new core_1.InternalError('ClickHouse not yet support exporting data to parquet file for caching datasets feature');
        });
    }
    checkProfileExist(profileName) {
        if (!this.clientMapping.has(profileName)) {
            throw new core_1.InternalError(`Profile instance ${profileName} not found`);
        }
    }
    convertToQueryParams(bind) {
        // find the param name from named placeholder from prepare method
        const pattern = /{(\w+):\w+}/;
        const params = {};
        bind.forEach((value, key) => {
            // get the param name from key. e.g: "{p1:Unit8}" => "p1"
            const paramName = key.match(pattern)[1];
            params[paramName] = value;
        });
        return params;
    }
};
ClickHouseDataSource = tslib_1.__decorate([
    (0, core_1.VulcanExtensionId)('clickhouse')
], ClickHouseDataSource);
exports.ClickHouseDataSource = ClickHouseDataSource;
//# sourceMappingURL=clickHouseDataSource.js.map